(* 	      On Graph Rewriting, Reduction and Evaluation.
 *			      Ian Zerny
 *	                   Copyright 2009
 *                GPLv3, see LICENSE.txt for details
 *
 *		      www.zerny.dk/on-graph-rewriting.html
 *
 *			Version: November 2009
 *)

(* Implementation of bracket abstraction
 * mapping a lambda-term to a lambda-term using only the SKI combinators.
 *)
structure BA = struct

datatype t
  = VAR of string
  | LAM of string * t
  | APP of t * t

(*  ba : t -> t *)
fun ba (VAR x)
    = (VAR x)
  | ba (LAM (x, e))
    = bavar (x, e)
  | ba (APP (e0, e1))
    = APP (ba e0, ba e1)

(*  bavar : string * t -> t *)
and bavar (x, e)
    = if isfv (x, e)
      then (case e
             of VAR y        => VAR "I"
              | LAM (y, e0)  => bavar (x, bavar (y, e0))
              | APP (e0, e1) => APP (APP (VAR "S", bavar (x, e0)),
                                     bavar (x, e1)))
      else APP (VAR "K", ba e)

(*  isfv : string * t -> bool *)
and isfv (x, VAR y)
    = (x = y)
  | isfv (x, LAM (y, e))
    = (x <> y) andalso
      isfv (x, e)
  | isfv (x, APP (e0, e1))
    = isfv (x, e0) orelse 
      isfv (x, e1)

end;
