(* 	      On Graph Rewriting, Reduction and Evaluation.
 *			      Ian Zerny
 *	                   Copyright 2009
 *                GPLv3, see LICENSE.txt for details
 *
 *		      www.zerny.dk/on-graph-rewriting.html
 *
 *			Version: November 2009
 *)

(* Simple evaluator for lambda-terms with integers
 * using higher-order abstract syntax. *)
structure HO
= struct

datatype t
  = LIT of int
  | LAM of t -> t
  | APP of t * t

exception NON_APPLICABLE_VALUE
exception NON_INTEGER_VALUE

val zro = LIT 0
val scc = LAM (fn (LIT i) => LIT (i+1)
                | _       => raise NON_INTEGER_VALUE)

val I = LAM (fn x => x)

val K = LAM (fn x =>
          LAM (fn y => x))

val S = LAM (fn x =>
          LAM (fn y =>
            LAM (fn z => APP (APP (x, z), APP (y, z)))))

(*  eval : t -> t *)
fun eval (LIT i) = LIT i
  | eval (LAM f) = LAM f
  | eval (APP (f, a))
    = (case eval f
        of LAM f => eval (f (eval a))
         | _ => raise NON_APPLICABLE_VALUE)

(*  c2n : t -> int *)
fun c2n c = eval (APP (APP (c, scc), zro))

end;
